#!/usr/bin/env python3
import sys
import ctypes as ct
from bcc import BPF

title = "n_tty_write"

try:
    if sys.argv[1]:
        tty_inode = sys.argv[1]
except IndexError as e:
    raise ValueError("No tty specified to snoop on") from e
txt = "#define NTW_TTY_INODE {}".format(tty_inode)

txt += """

#include <linux/tty.h>
#include <linux/sched.h>

struct write_fact {
    u64 pid;
    size_t nr;
    char comm[TASK_COMM_LEN];
    char name[64];
    char buf[64];
};

BPF_PERF_OUTPUT(oq);

int kprobe__n_tty_write(struct pt_regs *ctx,
        struct tty_struct *tty, struct file *file, const unsigned char *buf, size_t nr)
{
    if (file->f_inode->i_ino != NTW_TTY_INODE)
        return 0;

    struct write_fact wf = {};
    bpf_get_current_comm(&wf.comm, sizeof(wf.comm));
    wf.pid = bpf_get_current_pid_tgid() >> 32;
    wf.nr = nr;
    bpf_probe_read(&wf.name, sizeof(wf.name), tty->name);
    bpf_probe_read(&wf.buf, nr < sizeof(buf) ? nr : sizeof(buf), buf);

    oq.perf_submit(ctx, &wf, sizeof(wf));
    return 0;
}

"""

b = BPF(text=txt)

# Let's define the output data structure (write_fact)
# in Python to be able to interpret it as a blob.
TASK_COMM_LEN = 16    # linux/sched.h
class Record(ct.Structure):
    _fields_ = [("pid", ct.c_ulonglong),
                ("nr", ct.c_ulonglong),
                ("comm", ct.c_char * TASK_COMM_LEN),
                ("name", ct.c_char * 64),
                ("buf", ct.c_char * 64)]

def print_event(cpu, data, size):
    event = ct.cast(data, ct.POINTER(Record)).contents
    print("%-6d %-19s %s(%s, %s, %d)" %
            (event.pid, event.comm, title, event.name, event.buf, event.nr))

# loop with callback to print_event
b["oq"].open_perf_buffer(print_event)
while 1:
    b.perf_buffer_poll()

